#include "corpinc.h"

void setup_sgml(void);
BOOL get_sgml(void);
int IsDelimiter(void);
int RecognizeDelimiter(void);
BOOL TestContext(int);
BOOL is_nc(char);
void skip_sgml(int,int);
void get_sgml_name(char*);
void get_sgml_val(char*);
void s_adv(int);
void s_read(int,char*);                         
char s_get(int);
void ClassifyEntity(void);
void ClassifyCref(void);
int ClassifyPOS(char*);
void skip_blanks(void);
void s_init(void);
BOOL bWordBreak(char,char);
BOOL SGMLComp(const char*,const char*);
unsigned char WordFold(unsigned char);
enum _st ch_st(unsigned char);
enum _st el_st(char*);
enum _st cr_st(char*);

long pos; WORDNO nwo;
int context;
char thischar;
struct ientry indx;
FILE *fSource; 


static BOOL bAttList;
char sElement[MAXNAME+1];

struct sgml record;
char lemma[MAXWORD];


void setup_sgml(void)
{                              
	s_init();  nwo=0;
	context=CO_CON;
	bAttList=FALSE;
/* Search for start of document instance */
	record.id=HDRPOS;
	return;
}	

BOOL get_sgml(void)
{
/* Fill record with details of the next SGML item as follows.
SG_STARTTAG		A start tag (starts <)
SG_ENDTAG		An end tag (starts </)
SG_CHAR			A character in context
SG_CENT			A character entity
SG_PENT			A part-of-speech entity
*/	        
	static long oldpos; BOOL bCh=TRUE;           
	if (bAttList)
	{
		skip_blanks();
		get_sgml_name(record.name);
		switch (IsDelimiter())
		{
			case CH_VI:
				get_sgml_val(record.val);
				record.type=SG_ATT;
				record.pos=oldpos;
				break;
			case CH_TAGC:
				if (record.name[0]==0)
					{
					bAttList=FALSE;
					break;
					}
				strcpy(record.val,record.name);
				strcpy(record.name,"pos");
				record.type=SG_ATT;
				record.pos=oldpos;
				bAttList=FALSE;
				return TRUE;
			default:
				bAttList=FALSE;
		}
		if (bAttList) return TRUE;
	}						
	oldpos=pos;	
	switch (IsDelimiter())
	{
	case CH_STAGO:
		get_sgml_name(record.name);
		record.type=SG_STARTTAG;
		record.pos=oldpos;
		bAttList=TRUE;
		strcpy(sElement,record.name);
		break;
	case CH_ETAGO:	
		get_sgml_name(record.name);
		record.type=SG_ENDTAG;
		record.pos=oldpos;
		skip_sgml(CO_TAG,CH_TAGC);
		break;            
	case CH_ERO:
		get_sgml_name(record.name);
		ClassifyEntity();
		record.pos=oldpos;
		if (s_get(0)==';') s_adv(1);
		break;
	case CH_CRO:
		get_sgml_name(record.name);
		ClassifyCref();
		record.pos=oldpos;
		if (s_get(0)==';') s_adv(1);
		break;
	case CH_EOF:
		bCh=FALSE;	
		break;
	default:
		record.type=SG_CHAR;
		record.name[0]=thischar;
		record.pos=oldpos;
	}				
	return bCh;
}					

/* Context table for SGML as used here.             
Delimiters allowed in a TAG: LIT LITA TAGC VI
Delimiters allowed in a REF: REFC
Delimiters allowed in a LIT: LIT CRO ERO 
Delimiters allowed in a LITA: LITA CRO ERO 
Delimiters allowed in content: CRO ERO ETAGO STAGO
*/                  



#define NDELIMS 9
#define NCONTEXTS 5
char* delim[NDELIMS] = {"'","</","<",">","&#","&",";","\"","="};
BOOL bRec[NDELIMS][NCONTEXTS]=
	{
		{FALSE,FALSE,FALSE,TRUE,TRUE},
		{TRUE,FALSE,FALSE,FALSE,FALSE},
		{TRUE,FALSE,FALSE,FALSE,FALSE},
		{FALSE,FALSE,FALSE,FALSE,TRUE},
		{TRUE,FALSE,TRUE,TRUE,FALSE},
		{TRUE,FALSE,TRUE,TRUE,FALSE},
		{FALSE,TRUE,FALSE,FALSE,FALSE},
		{FALSE,FALSE,TRUE,FALSE,TRUE},
		{FALSE,FALSE,FALSE,FALSE,TRUE}
	};		

BOOL bContext[NDELIMS]={FALSE,TRUE,TRUE,FALSE,TRUE,TRUE,FALSE,FALSE,FALSE};

int IsDelimiter(void)
{                        
	int i;
	i=RecognizeDelimiter();
	s_adv(1);
	if (i==CH_CHAR||i==CH_EOF) return i;
	if (bRec[i][context])
		if (!TestContext(i)) return CH_CHAR;
	s_adv(strlen(delim[i])-1);
	return i;
}                       

static char test[3]; 

int RecognizeDelimiter(void)
{            
	int i,j; int b;
    s_read(3,test);
    thischar=test[0];
    for (i=0;i<NDELIMS;i++)   
    	{                               
    	j=0;
        while ((b=delim[i][j])!=0)
            if (test[j]!=delim[i][j]) break; else j++;
        if (!b) break;
        }
    if (i<NDELIMS) return i;
    if (test[0] != EOF) return CH_CHAR; else return CH_EOF;
}

BOOL TestContext(int i)
{
	if (bContext[i])
		return isalnum(test[strlen(delim[i])]); // NAME char
	else
		return TRUE;
}

BOOL is_nc(char c)			            
{
	return isalnum(c)||c=='-'||c=='.';
}	

void skip_sgml(int cont,int echar)
{
	int oldcontext=context; int i;
	context=cont;
	while (TRUE)
	{
	i=IsDelimiter();              
	if (i==echar||i==CH_EOF)
		{
		context=oldcontext;
		return;
		}
	switch (i)
	{
	case CH_LITA:
		skip_sgml(CO_LITA,CH_LITA);
		break;
	case CH_LIT:
		skip_sgml(CO_LIT,CH_LIT);
		break;
	case CH_STAGO:
		skip_sgml(CO_TAG,CH_TAGC);
		break;
	case CH_ETAGO:
		skip_sgml(CO_TAG,CH_TAGC);
		break;
	case CH_ERO:
		skip_sgml(CO_REF,CH_REFC);
		break;
	case CH_CRO:
		skip_sgml(CO_REF,CH_REFC);
		break;             
	default: 
		break;
	}
	}				
}		                        

void skip_blanks(void)
{
	while (isspace(s_get(0))) s_adv(1);
	return;
}

#define XTAG 30

int ClassifyPOS(char* s)
{
	int i, j;
	for (i=0;i<npos;i++)
		if (!strcmp(s,postab[i].name)) break;
	if (i<npos)      
  		return i;
	for (j=0;j<nukpos;j++) {
		if (!strcmp(s,ukpos[j])) break;
	}
	if (j<nukpos) return XTAG;
	corperr("Unrecognized part-of-speech %s\n",s);
	if (nukpos < MAXENTS) 
		strcpy(ukpos[nukpos++],s);
	return XTAG;
}

void ClassifyEntity(void)
{
	int i;
	for (i=0;i<npos;i++)
		if (!strcmp(record.name,postab[i].name)) break;
	if (i<npos)      
		{
        record.type=SG_PREF;
        record.id=i;
        }
    else     
    {
    for (i=0;i<nctab;i++)
    	if (!strcmp(record.name,ctab[i].cname)) break;

    if (i<nctab && ctab[i].val>=0 && ctab[i].val < 256)
    {
    	record.type=SG_CHAR;	
    	record.name[0]=ctab[i].val;
    }
    else if (i<nctab)	
    	record.type=SG_CREF;
    else
      {
	for (i=0;i<nspec;i++)
	  if (!strcmp(record.name,specs[i].name)) break;
        if (i<nspec)
	  		record.type=SG_CREF;
	else {
		int j; 
		record.type=SG_CREF;
		for (j=0;j<nukents;j++) {
			if (!strcmp(record.name,uketab[j])) break;
		}
		if (j < nukents) return;
		if (nukents < MAXGENTS) {
			strcpy(uketab[nukents],record.name);
			nukents++;
		}
 	  corperr("Unknown entity reference %s\n",record.name);
      }
    }}	
    return;
}                               

void ClassifyCref(void)
{
	int i=0;
	char *s=record.name;    	    
	while (*s != 0)
		i=10*i+*(s++)-'0';
	record.type=SG_CHAR;
	record.name[0]=i;
	return;
}
		
void get_sgml_name(char *s)			
{                         
	char c;
	while (TRUE)
	{
	s_read(1,&c);       
	if (is_nc(c)||c=='.'||c=='-')
		{
		*(s++)=c;
	    s_adv(1);
		}
	else
		break;
	}	
	*s=0;
	return;
}			    

void get_sgml_val(char *s)
{
	char c; char* s0=s; BOOL bErr=FALSE;
	s_read(1,&c);
	switch (c)
	{
	case '\"':
		while (TRUE)
		{
		s_adv(1);
		s_read(1,&c);
		if (c=='\"') break;
		if (!bErr) *(s++)=c;                      
		if (!bErr && s-s0 >= MAXATTVAL-1)
			{
			*s=0;
			corperr("Attribute value \"%s\" too long\n",s0);
			bErr=TRUE;
			}
		}
		s_adv(1);
		*s=0;
		break;
	case '\'':
		while (TRUE)
		{
		s_adv(1);
		s_read(1,&c);
		if (c=='\'') break;
		if (!bErr) *(s++)=c;
		if (!bErr && s-s0 >= MAXATTVAL-1)
			{
			*s=0;
			corperr("Attribute value '%s' too long\n",s0);
			bErr=TRUE;
			}
		}
		s_adv(1);
		*s=0;
		break;    
	default:
		get_sgml_name(s);
	}
	return;
}

void s_read(int n,char *buff)
{   
	int i;
	for (i=0;i<n;i++)
		buff[i]=s_get(i);
	return;
}	
               
#define NBUFF 128
char xbuff[NBUFF]; char * ibuff;
int ncur,nlast; 
int nbuff;              

void s_init()
{
	pos=0;
	ncur=0;
	nlast=0;
	nbuff=NBUFF;
	ibuff=xbuff;
	return;
}	

void s_init_string(char* s,long p)
{
	pos=p;
	ncur=0;
	nlast=strlen(s);
	nbuff=0;
	ibuff=s;
	return;
}	

char s_get(int i)
{               
	int j;
	j=ncur+i;
	if (nbuff > 0 && j>=nbuff) j-=nbuff;
	if (j != nlast)
		return ibuff[j];
	else if (nbuff==0) 
		return EOF;
	else
		{   
		char c;
		c=fgetc(fSource);		
		ibuff[j]=c;
		nlast = ((nbuff==0 || nlast<nbuff-1)?nlast+1:0);
		return c;
		}
}

void s_adv(int n)
{      
	ncur += n;
	pos += n;
	if (nbuff > 0 && ncur >= nbuff) ncur-=nbuff;
	return;
}

enum _st el_st(char* s)
{
	int i;
	for (i=0;i<nelts;i++) {
		if (SGMLComp(eltab[i].name,s)) break;
	}
	if (i < nelts) return eltab[i].st;
	return ST_SPACE;
}

enum _st cr_st(char* s)
{
	return ST_CHAR;
}



enum _st ch_st(unsigned char c)
{
	return lextab[c];
}

unsigned char WordFold(unsigned char c)
{
	int i;
	if (isascii(c)) return tolower(c);
	for (i=0;i<ncase;i++) 
		if (lctab[i].u==c) return lctab[i].l;
	return c;
}

void init_sgml()
{
	context=CO_CON;
	bAttList=FALSE;
}

int IsFrob(char* w)
{
	int i;
	for (i=0;i<nfrobs;i++)
		if (!strcmp(w,frobs[i])) return i;
	return -1;
}

BOOL IsWTag(const char* s)
{
	int i=0;
	for (i=0;i<nwt;i++) 
		if (SGMLComp(s,wtags[i].name)) return TRUE;
	return FALSE;
}

BOOL IsPosTag(const char* s,const char* a)
{
	int i=0;
	for (i=0;i<nwt;i++) 
		if (SGMLComp(s,wtags[i].name) && SGMLComp(a,wtags[i].pos)) return TRUE;
	return FALSE;
}

BOOL IsLemmaTag(const char* s,const char* a)
{
	int i=0;
	for (i=0;i<nlt;i++) 
		if (SGMLComp(s,ltags[i].name) && SGMLComp(a,ltags[i].pos)) return TRUE;
	return FALSE;
}


void Attval(char* s)
{
	char* s1=s;
	while (*s) {
		if (*s=='&')
		{
			char* t; int i;
			s++;
			t=s;
			while (*t && *t != ';') t++;
			*t=0;
			for (i=0;i<nctab;i++) {
    			if (!strcmp(s,ctab[i].cname)) break;
			}
			s=t+1;
			if (i==nctab) continue;
			*s1++=ctab[i].val;
		}
		else 
			*s1++=*s++;
	}
	*s1=0;
}


int hashlfn(const char* s)
{
	int i, h; h=0;
	for (i=0;i<4;i++) {
		if (s[i]==0) break;
		h+=s[i]<<(5*i);
	}
	h &= 0xffff;
	return h;
}

struct _lt {
	char w[MAXWORD];
	char lw[MAXWORD];
	int pos;
	int ptr;
};

#define MAXLTAB 100000

struct _lt ltab[MAXLTAB];
int hlt[0x10000];

int lts; 

void InitLTab()
{
	int i;
	lts=0;
	for (i=0;i<0x10000;i++)
		hlt[i]=-1;
}

struct _lt* LTLookup(const char* w,int pos,int k)
{
	int h=hashlfn(w);
	int j=hlt[h];
	while (j != -1) {
		if (!strcmp(ltab[j].w,w) && ltab[j].pos==pos) {
			if (k==0) return &ltab[j];
			k--;
		}
		j=ltab[j].ptr;
	}
	return NULL;
}


int AddLemma(const char* w,int pos,const char* lw)
{
	int k; struct _lt* lt; int h;
	if (pos==HDRPOS) return 0;
	for (k=0;;k++) {
		lt=LTLookup(w,pos,k);
		if (lt==NULL) break;
		if (!strcmp(lw,lt->lw)) return k;
	}
	h=hashlfn(w);
	strcpy(ltab[lts].w,w);
	strcpy(ltab[lts].lw,lw);
	ltab[lts].pos=pos;
	ltab[lts].ptr = -1;
	if (hlt[h]==-1) hlt[h]=lts;
	else {
		int j=hlt[h];
		while (ltab[j].ptr != -1)
			j=ltab[j].ptr;
		ltab[j].ptr=lts;
	}
	lts++;
	return k;
}

void LemmaSave()
{
	char path[_MAX_PATH]; int i;FILE *f;
	strcpy(path,etcpath);
	strcat(path,"lemmtab");
	f=fopen(path,"w"); 
	for (i=0;i<lts;i++) {
		fprintf(f,"%s\t%d\t%s\n",ltab[i].w,ltab[i].pos,ltab[i].lw);
	}
	fclose(f);
	corperr("Lemma table size is %d\n",lts);
}

#define MAXLINE 1000

void LemmaLoad()
{
	FILE* f; char path[_MAX_PATH]; char line[MAXLINE];
	strcpy(path,etcpath);
	strcat(path,"lemmtab");
	f=fopen(path,"r");
	if (f==NULL) return;
	InitLTab();
	while (fgets(line,MAXLINE,f)) {
		char w[MAXWORD]; char lw[MAXWORD];
		int pos; 
		sscanf(line,"%[^\t]\t%d\t%[^\n] ",w,&pos,lw);
		AddLemma(w,pos,lw);
	}
}

BOOL lemmata1(char* w,int pos,char* lw)
{
	struct _lt* lt=LTLookup(w,pos%10000,pos/10000);
	if (lt==NULL) return FALSE;
	strcpy(lw,lt->lw);
	return TRUE;
}
