/*
 * $Id: corppq.c,v 1.2 1996/05/12 23:36:15 scotty Rel $
 *
 * $Log: corppq.c,v $
 * Revision 1.2  1996/05/12 23:36:15  scotty
 * v927 fixes from Tony.
 *
 * Revision 1.1  1996/02/15 01:29:28  scotty
 * Initial revision
 *
 */
#include "corpinc.h"

void      Normalize(char *, char *);

static char rcsid[] = "$Id: corppq.c,v 1.2 1996/05/12 23:36:15 scotty Rel $";

#define SERVER 1

#define MAXQSTRING   50
#define MAXQTOK     500
#define MAXPD       100

#define TK_INFIX   1
#define TK_PREFIX  2
#define TK_OP      3
#define TK_CP      4
#define TK_OB      5
#define TK_CB      6
#define TK_OT      7
#define TK_CT      8
#define TK_NAME    9
#define TK_VAR    10
#define TK_LOC    11
#define TK_NVAR   12


struct loc {
   sara_ulong  code;
   WORDNO      w0;
   WORDNO      nw;
};


struct {
   int         ty;
   char        cha;
   char        str[MAXQSTRING];
   struct loc  l;
   int         num;
} tok[MAXQTOK];


int   nt;


struct {
   int   val;
   int   scope;
} stack[MAXPD];

int   st;

struct {
   BOOL      bHeap;
   int       arg;
} argstack[MAXPD];

int   nargstack;


union {
   struct {
      char fn;
      int  arity;
   } op;
   struct {
      BOOL bHeap;
      int  val;
   } arg;
} heap[100];

int   ht;


int       TokeniseQuery(char *);
BOOL      GetToken(char **, int *);
BOOL      bOKName(unsigned char);
void      GetName(char **, char *);
void      GetString(char **, char *, char);
BOOL      ParseQuery(void);
void      pushop(int, int);
int       pop(int);
void      pusharg(int);
BOOL      bPrec(int, int);
int       ParseTag(int *);
int       prec(int);
void      PrintQuery(BOOL, int, int);
struct stream *BuildQuery(BOOL, int);
int       optype(char);
int       GetInt(char *);
void      GetLoc(char **, struct loc *);
void      Defrob(char *, char *);
char     *GetWord(char **);
BOOL      LetterChar(unsigned char);
BOOL	SpaceChar(unsigned char);
int       enclitic(char *);

extern int ClassifyPOS(char*);

BOOL      bCase;
BOOL	  bLemma;

#if !defined(SERVER)

void FullQuery(void)
{
   char           b[MAXQSTRING * MAXQTOK];
   int            i;
   int            n;

   struct stream *s;

   NewQuery();
   FreeStreams();

   printf(">");
   scanf("%[^\n]", b);

   nt = TokeniseQuery(b);

   if (bDebug)
      for (i = 0; i < nt; i++)
         printf("%5d%2c%s\n", tok[i].ty, tok[i].cha, tok[i].str);

   if (!ParseQuery())
      return;

   if (bDebug)
      PrintQuery(argstack[0].bHeap, argstack[0].arg, 0);

   s = BuildQuery(argstack[0].bHeap, argstack[0].arg);

   if (s == NULL)
      return;

   qs = *s;

   n = CountStream(&qs);

   printf("%s\n", qs.word);

   printf("%i instance%s found\n", n, (n != 1 ? "s" : ""));

   if (n == 0)
      qs.type = 0;

   return;
}


#endif /* ! SERVER */


int TokeniseQuery(char *qy)
{
   int   i = 0;
   char *q = qy;

   while (GetToken(&q, &i))
      ;

   return i;
}


BOOL GetToken(char **qy, int *i)
{
   while (isspace(**qy))
      (*qy)++;

   switch (**qy) {
    case 0:
      return FALSE;

    case '_':
      tok[*i].ty = TK_VAR;
      (*qy)++;

      if (!isalnum(**qy)) {
         (*i)++;
         return TRUE;
      }

      tok[*i].ty = TK_NVAR;
      GetName(qy, tok[*i].str);
      (*i)++;

      return TRUE;

    case '|':                   /* FALLTHROUGH */
    case '*':                   /* FALLTHROUGH */
    case '=':                   /* FALLTHROUGH */
    case '/':                   /* FALLTHROUGH */
    case '#':
      tok[*i].ty = TK_INFIX;
      tok[*i].cha = **qy;
      (*qy)++;
      (*i)++;
      return TRUE;

    case '(':
      tok[*i].ty = TK_OP;
      (*qy)++;
      (*i)++;

      return TRUE;

    case ')':
      tok[*i].ty = TK_CP;
      (*qy)++;
      (*i)++;
      return TRUE;

    case '{':
      tok[*i].ty = TK_OB;
      GetString(qy, tok[*i].str, '}');
      (*i)++;
      return TRUE;

    case '<':
      tok[*i].ty = TK_OT;
      (*qy)++;
      (*i)++;
      return TRUE;

    case '>':
      tok[*i].ty = TK_CT;
      (*qy)++;
      (*i)++;
      return TRUE;

    case '!':                   /* FALLTHROUGH */
    case ':':                   /* FALLTHROUGH */
    case ';':                   /* FALLTHROUGH */
    case '$':                   /* FALLTHROUGH */
	case '^':					/* FALLTHROUGH */
	case '~':
    case '?':                   /* FALLTHROUGH */
    case '@':
      tok[*i].ty = TK_PREFIX;
      tok[*i].cha = **qy;
      (*i)++;
      (*qy)++;
      return TRUE;

    case '\"':                  /* " for emacs */
      tok[*i].ty = TK_NAME;
      GetString(qy, tok[*i].str, **qy);
      (*i)++;
      return TRUE;

    case '\'':
      {
         char  b1[MAXPHRASE];
         char     b2[MAXPHRASE];
         char *s;

         GetString(qy, b1, **qy);
         Defrob(b1, b2);
         s = b2;

         if (s[0] == 0)
            return TRUE;

         do {
            if (!strcmp(s, "_")) {
               tok[*i].ty = TK_VAR;
            } else {
               tok[*i].ty = TK_NAME;
               strcpy(tok[*i].str, s);
            }
            (*i)++;
            s += strlen(s) + 1;
         } while (s[0] != 0);

         return TRUE;
      }

    case '&':
      tok[*i].ty = TK_LOC;
      GetLoc(qy, &tok[*i].l);
      (*i)++;
      return TRUE;


    default:
      if (bOKName(**qy)) {
         tok[*i].ty = TK_NAME;
         GetName(qy, tok[*i].str);
         (*i)++;
         return TRUE;
      } else {
         char  c[2];

         c[0] = **qy;
         c[1] = 0;

         corperr("Unrecognised character %s in query", c);

         return FALSE;
      }
   }
}


void GetLoc(char **q, struct loc *l)
{
   sara_ulong  l1;
   int         n;

   l1 = (*q[1] << 16) + (*q[2] << 8) + *q[3];

   *q += 4;

   l->code = l1;
   l->w0 = -1;
   l->nw = 1;

   if (**q != ':')
      return;

   n = 0;

   while (isdigit(**q))
      n = 10 * n + (*((*q)++) - '0');

   l->w0 = n;

   if (**q != '-')
      return;

   n = 0;

   while (isdigit(**q))
      n = 10 * n + (*((*q)++) - '0');

   l->nw = n;

   return;
}


BOOL bOKName(unsigned char c)
{
   return isalnum(c) || ispunct(c) || c > 128;
}


void  GetName(char **q, char *buff)
{
   int   i = 0;

   if (ispunct(**q)) {
      buff[0] = **q;
      buff[1] = 0;
      (*q)++;
      return;
   }
	   
   while (isalnum(**q) || **q == '-' || **q=='.' || (**q=='\\' && *(*q)++)) {
      buff[i++] = *((*q)++);

      if (i > MAXQSTRING - 2) {
         buff[i] = 0;
         corperr("Query string %s too long", buff);
         while (isalpha(**q))
            (*q)++;
         return;
      }
   }

   buff[i] = 0;

   return;
}


void GetString(char **q, char *buff, char end)
{
   int   i = 0;

   (*q)++;

   while (TRUE) {
      switch (**q) {

       case '\\':
         (*q)++;
         if (**q == 0)
            return;

         buff[i++] = *((*q)++);
         break;

       case '"':                /* FALLTHROUGH */
       case '\'':               /* FALLTHROUGH */
       case '}':
         if (**q == end) {
            (*q)++;
            buff[i] = 0;
            return;
         }
         /* FALLTHROUGH */
       default:
         buff[i++] = *((*q)++);
         break;
      }

      if (i > MAXQSTRING - 2) {
         buff[i] = 0;
         corperr("Query string %s too long", buff);
         while (**q != 0 && **q != '"')
            (*q)++;
         return;
      }
   }
}


int   nscope;


BOOL ParseQuery(void)
{
   int   i,
         h;
   int   currop;

   currop = -1;
   nscope = 0;

   ht = st = nargstack = 0;

   if (nt == 0)
      return FALSE;

   for (i = 0; i < nt; i++)
      switch (tok[i].ty) {

       case TK_NAME:            /* FALLTHROUGH */
       case TK_VAR:             /* FALLTHROUGH */
       case TK_NVAR:            /* FALLTHROUGH */
       case TK_OB:
//         if (bPrec(currop, -1) && nargstack == nscope) {
//            pusharg(i);
//            break;
//         }
		 if (currop >= 0 && tok[currop].ty==TK_PREFIX && nargstack==nscope) {
			   pusharg(i);
			   break;
		   }
		 if (currop >= 0 && tok[currop].ty==TK_INFIX && nargstack-1==nscope) {
			   pusharg(i);
               pushop(currop, nscope);
               nscope = nargstack-1;
               currop = -1;
			   break;
		   }
         while (bPrec(currop, -1))
            currop = pop(currop);

         if (currop != -1) {
            pushop(currop, nscope);
            nscope = nargstack;
            currop = -1;
         }
         pusharg(i);
         break;

       case TK_INFIX:

       reop:

         if (currop >= 0 && tok[i].cha == tok[currop].cha)
            break;

         if (bPrec(i, currop)) {
            pushop(currop, nscope);
            nscope = nargstack - 1;
            currop = i;
         } else {
            currop = pop(currop);
            goto reop;
         }
         break;

       case TK_PREFIX:
         pushop(currop, nscope);
         currop = i;
         nscope = nargstack;
         break;

       case TK_OP:
         pushop(currop, nscope);
         currop = -2;
         nscope = nargstack;
         break;

       case TK_CP:
         while (currop != -2) {
            currop = pop(currop);
            if (currop == -3) {
               corperr("Syntax error", NULL);
               return FALSE;
            }
         }

/*
         currop=pop(currop);
*/
         currop = -1;
         break;

       case TK_OT:
         h = ParseTag(&i);
         if (h < 0) {
            corperr("Syntax error", NULL);
            return FALSE;
         }
         if (currop != -1) {
            pushop(currop, nscope);
            nscope = nargstack;
            currop = -1;
         }
         argstack[nargstack].arg = h;
         argstack[nargstack++].bHeap = TRUE;
         break;
      }

   do {
      currop = pop(currop);
      if (currop == -2) {
         corperr("Syntax error", NULL);
         return FALSE;
      }
   } while (currop != -3);

   return TRUE;
}


void pushop(int i, int n)
{
   stack[st].scope = n;
   stack[st++].val = i;
   return;
}


int pop(int i)
{
   int   h = ht;

   if (st < 0)
      return -3;

   if (i == -2) {
      st--;
      nscope = stack[st].scope;
      return stack[st].val;
   }
   if (i == -1) {
      if (nargstack == nscope + 1)
         goto pf;
      heap[ht].op.fn = '+';
   } else
      heap[ht].op.fn = tok[i].cha;

   heap[ht++].op.arity = nargstack - nscope;

   for (i = nscope; i < nargstack; i++) {
      if (argstack[i].bHeap) {
         heap[ht].arg.bHeap = TRUE;
         heap[ht++].arg.val = argstack[i].arg;
      } else {
         heap[ht].arg.bHeap = FALSE;
         heap[ht++].arg.val = argstack[i].arg;
      }
   }

   argstack[nscope].bHeap = TRUE;
   argstack[nscope].arg = h;
   nargstack = nscope + 1;

 pf:

   if (--st < 0)
      goto pf1;

   nscope = stack[st].scope;
   return stack[st].val;

 pf1:

   st = 0;
   nscope = 0;
   return -3;
}


void pusharg(int i)
{
   argstack[nargstack].arg = i;
   argstack[nargstack++].bHeap = FALSE;
   return;
}


BOOL bPrec(int new, int curr)
{
   int   i1 = prec(new);
   int   i2 = prec(curr);

   return i1 > i2;
}


int prec(int i)
{
   if (i <= -2)
      return 0;

   if (i == -1)
      return 100;

   switch (tok[i].cha) {

    case '|':
      return 50;

    case '&':
      return 40;

    case '*':
      return 30;

    case '#':
      return 25;

    case '/':
      return 20;

    case '=':
      return 130;

    case ':':                   /* FALLTHROUGH */
    case ';':                   /* FALLTHROUGH */
    case '$':                   /* FALLTHROUGH */
	case '^':
	case '~':
    case '@':
      return 120;

    case '!':
      return 110;

    case '?':
      return 105;
   }

   /*
    * for the weird case, a.k.a. "None of the above"
    */
   return 0;
}


int ParseTag(int *i)
{
   BOOL  bEnd = FALSE;
   int   h = ht;

   (*i)++;

   if (tok[*i].ty == TK_INFIX && tok[*i].cha == '/') {
      bEnd = TRUE;
      (*i)++;
   }
   heap[ht++].op.fn = bEnd ? '>' : '<';

   if (tok[*i].ty != TK_NAME)
      return -1;

   heap[ht].arg.val = *i;
   heap[ht++].arg.bHeap = FALSE;

   (*i)++;

   while (tok[*i].ty != TK_CT) {
      if (tok[*i + 1].ty != TK_INFIX || tok[*i + 1].cha != '=')
         return -1;

      heap[ht].arg.val = *i;
      heap[ht++].arg.bHeap = FALSE;
      heap[ht].arg.val = *i + 2;
      heap[ht++].arg.bHeap = FALSE;

      *i += 3;
   }

   heap[h].op.arity = ht - h - 1;
   return h;
}


#if !defined(SERVER)


void PrintQuery(BOOL bH, int h, int off)
{
   if (bH) {
      int   n = heap[h].op.arity;
      int   i;

      printf("%*s%c\n", off, "", heap[h].op.fn);

      for (i = 0; i < n; i++)
         PrintQuery(heap[h + 1 + i].arg.bHeap,
                    heap[h + 1 + i].arg.val, off + 6);
   } else
      printf("%*s%s\n", off, "", tok[h].str);

   return;
}


#endif /* ! SERVER */


void FlattenStream(struct stream *s)
{
   struct stream *t[MAXOPS];
   int            i,
                  j,
                  k;

   k = 0;
   for (i = 0; i < s->ops; i++) {
      if (s->op[i]->type == TY_SEQ && s->type == TY_SEQ) {
         for (j = 0; j < s->op[i]->ops; j++)
            t[k++] = s->op[i]->op[j];
      } else
         t[k++] = s->op[i];
   }
   s->ops = k;

   for (i = 0; i < k; i++)
      s->op[i] = t[i];
}


struct stream *BuildQuery(BOOL bH, int h)
{
   if (bH) {
      int            n = heap[h].op.arity;
      int            h1;
      int            tn,
                     tp;
      int            jj;
      struct stream *s,
                    *s1,
                    *s2,
                    *s3;
      int            j;
      char          *attnames[MAXATTS];
      char          *attvals[MAXATTS];
      int            nat;

      switch (heap[h].op.fn) {

       case '|':                /* FALLTHROUGH */
       case '+':                /* FALLTHROUGH */
       case '*':                /* FALLTHROUGH */
       case '#':
         s = GrabStream();
         jj = 0;
         s->type = optype(heap[h].op.fn);

         for (j = 0; j < n; j++) {
            struct stream *ss = BuildQuery(heap[h + 1 + j].arg.bHeap,
                                           heap[h + 1 + j].arg.val);

            if (ss == NULL && heap[h].op.fn != '|')
               return NULL;
            if (ss != NULL)
               s->op[jj++] = ss;
            if (jj >= MAXOPS - 1)
               {
               bStreamErr=TRUE;
               return NULL;
               }
         }
         s->ops = jj;
         FlattenStream(s);
         return s;

       case '!':
         s = GrabStream();
         s->type = TY_NEG;
         s->ops = 1;
         s->op[0] = BuildQuery(heap[h + 1].arg.bHeap, heap[h + 1].arg.val);
         return s;

       case ':':
         return MakeFileStream(tok[heap[h + 1].arg.val].str);

       case ';':
         return MakeSCStream(tok[heap[h + 1].arg.val].str);

       case '$':
         bCase = TRUE;
         s = BuildQuery(heap[h + 1].arg.bHeap, heap[h + 1].arg.val);
         bCase = FALSE;
         return s;

       case '@':
         bHdr = TRUE;
         s = BuildQuery(heap[h + 1].arg.bHeap, heap[h + 1].arg.val);
         bHdr = FALSE;
         return s;

	   case '^':
         return MakeLemmaStream(tok[heap[h + 1].arg.val].str);

	   case '~':
         return MakeFormStream(tok[heap[h + 1].arg.val].str);

       case '/':
         s = BuildQuery(heap[h + 1].arg.bHeap, heap[h + 1].arg.val);
         if (s == NULL)
            return NULL;

         if (heap[h + 2].arg.bHeap) {
            int   i,
                  et;

            h1 = heap[h + 2].arg.val;

            switch (heap[h1].op.fn) {
             case '<':
               et = LG_ELEMENT;

               for (i = 0; i < nelts; i++)
                  if (SGMLComp(eltab[i].name, tok[heap[h1 + 1].arg.val].str))
                     break;

               if (i < nelts)
                  et = eltab[i].eltype;

               tp = LT_SGML;

               switch (et) {

                case LG_ELEMENT:
                  s1 = BuildQuery(TRUE, h1);                     
                  if (s1==NULL) return NULL;
                  s2 = MakeAtomStream(tok[heap[h1 + 1].arg.val].str,
                                      0, ID_TAGE);
                  break;

                case LG_SENT:
                  s1 = BuildQuery(TRUE, h1);
                  if (s1==NULL) return NULL;
                  s2 = MakeAtomStream(tok[heap[h1 + 1].arg.val].str, 0, ID_TAGO);
                  break;

                case LG_EMPTY:
                  corperr("%s is empty", eltab[i].name);
                  return NULL;
               }
               break;

             case '>':
               s1 = MakeAtomStream(tok[heap[h1 + 1].arg.val].str, 0, ID_TAGE);
               s2 = MakeAtomStream(tok[heap[h1 + 1].arg.val].str, 0, ID_TAGO);
               tp = LT_XSGML;
               break;
            }
         } else {
            tn = GetInt(tok[heap[h + 2].arg.val].str);
            tp = LT_INT;
         }

         if (s->type == TY_PROD || s->type == TY_BPROD) {
            s->limit1 = s1;
            s->limit2 = s2;
            s->limtype = tp;
            s->limint = tn;
            return s;
         } else {
            if (tp == LT_INT)
               return s;

            s3 = GrabStream();
            s3->type = TY_LIMIT;
            s3->op[0] = s;
            s3->ops = 1;
            s3->limit1 = s1;
            s3->limit2 = s2;
            s3->limtype = tp;
         }

         return s3;

       case '<':
         nat = (n - 1) / 2;

         for (j = 0; j < nat; j++) {
            attnames[j] = tok[heap[h + 2 + 2 * j].arg.val].str;
            /*
             * Making changes here to accommodate several values
             */
            if (heap[h + 2 + 2 * j + 1].arg.bHeap) {
               if (heap[h + 2 + 2 * j + 1].op.fn != '|')
                  return NULL;
            } else
               attvals[j] = tok[heap[h + 2 + 2 * j + 1].arg.val].str;
         }

         return MakeSGMLStream2(tok[heap[h + 1].arg.val].str, nat,
                                attnames, attvals);

       case '>':
         return MakeAtomStream(tok[heap[h + 1].arg.val].str, 0, ID_TAGE);

       case '?':
         s = BuildQuery(heap[h + 1].arg.bHeap, heap[h + 1].arg.val);
         if (s == NULL)
            return NULL;        /* not right */
         s->bOptional = TRUE;

         return s;

       case '=':
         {
            char *s = tok[heap[h + 1].arg.val].str;
            char *pos = tok[heap[h + 2].arg.val].str;
            int   i;
			if (tok[heap[h+1].arg.val].ty==TK_VAR) {
				return MakePosStream(pos);
			}

            if (!strcmp("HDR", pos))
               i = HDRPOS;
            else
               for (i = 0; i < npos; i++)
                  if (!strcmp(postab[i].name, pos))
                     break;

            if (i == npos)
               return NULL;

            return MakePosStream2(s, i + (c6?C6OFFSET:0));
         }

       default:
         return NULL;
      }
   } else if (tok[h].ty == TK_VAR) {
      struct stream *s = GrabStream();

      s->type = TY_ALL;
      s->wtype = ID_WORD;
      return s;
   } else if (tok[h].ty == TK_OB)
      return MakeRegExpStream(tok[h].str);
   else {
      struct stream *s = MakeAllPOSStream(tok[h].str, bHdr);

      if (s == NULL)
         return NULL;
      if (bCase)
         return MakeCaseFilter(s, tok[h].str);
      else
         return s;
   }
}


int optype(char c)
{
   switch (c) {
    case '|':
      return TY_OR;

    case '&':
      return TY_AND;

    case '+':
      return TY_SEQ;

    case '*':
      return TY_PROD;

    case '#':
      return TY_BPROD;
   }
   return TY_ATOM; // unused
}


int GetInt(char *s)
{
   int   k = 0;

   while (*s != 0) {
      k = k * 10 + *s - '0';
      s++;
   }
   return k;
}


sara_long Solve(char *name, char *cql, int *ntx)
{
   sara_long      n;
   struct stream *s;

   bInt = FALSE;

   NewQuery();
   FreeStreams();

   bStreamErr = FALSE;

   nt = TokeniseQuery(cql);
   if (nt >= MAXQTOK)
      return -1;

   if (!ParseQuery())
      return -1;

   s = BuildQuery(argstack[0].bHeap, argstack[0].arg);
   if (bStreamErr)
      return -4;
   if (s == NULL)
      return 0;

   qs = *s;

   n = SaveQueryFile(s, name, ntx);

   if (n <= 0)
      qs.type = 0;

   return n;
}


sara_long colloc_score(char *name, char *cql, int kwin1, int kwin2, int* ns)
{
   FILE             *f;
   struct locent1    l;
   sara_long         ks = 0;
   struct stream    *s;

/*   NewQuery();
   FreeStreams();

   nt = TokeniseQuery(cql);

   if (!ParseQuery())
      return -1;

   s = BuildQuery(argstack[0].bHeap, argstack[0].arg);
*/
   s=MakeLemmaStream(cql);
   if (s == NULL)
      return 0;

   RewindStream(s);
   XAdvanceStream(s);

   f = fopen(GetQPath(name), "r");

   while (fread(&l, sizeof(l), 1, f) == 1) {
/*      fix_locent1(&l, 1); */
      ks += xscore(&l, s, kwin1,kwin2);
   }

   fclose(f);
   RewindStream(s);
   *ns=0;
   while (XAdvanceStream(s)) (*ns)++;
   return ks;
}

BOOL      Frob(char **, char **);
char      frobuff[MAXPHRASE];


void Defrob(char *in, char *out)
{
   char *x;
   char *b = out;

   Normalize(in, frobuff);
   x = frobuff;
   while (Frob(&x, &b));
   b[0] = 0;
}


BOOL Frob(char **s, char **out)
{
#ifdef old_defrob
	int   i,
#endif
      int   j;
   int   ii = -1;

   if (**s == 0)
      return FALSE;
#ifdef old_defrob
   if (!bNoPos) {for (i = 0; i < nfrobs; i++) {
#if defined(__unix__)
      int   l = strncasecmp(*s, frobs[i], strlen(frobs[i]));
#else /* ! __unix__ */
      int   l = strnicmp(*s, frobs[i], strlen(frobs[i]));
#endif /* __unix__ */
      if (l == 0)
         ii = i;
   }}
#endif

   if (ii >= 0) {
      strncpy(*out, *s, strlen(frobs[ii]));
      (*out)[strlen(frobs[ii])] = 0;
      (*s) += strlen(frobs[ii]) + 1;
      (*out) += strlen(*out) + 1;
      return TRUE;
   }

   j = 0;

   while (*(*s + j) && *(*s + j) != ' ') {
      *(*out + j) = *(*s + j);
      j++;
   }

   *(*out + j) = 0;
   (*s) += j + 1;
   (*out) += j + 1;

   return TRUE;
}

#define PUNCT " .,;:"

void Normalize(char *in, char *out)
{
   char *s;
   int   i;
   char *p = in;

   out[0] = 0;
   while (s = GetWord(&p)) {

      int   n;
		if (bNoPos) strcat(out,s);

      else if (n = enclitic(s)) {
         strncat(out, s, n);
         strcat(out, " ");
         strcat(out, s + n);
      } else {
         for (i = 0; i < nmul; i++)
            if (!strcmp(s, mul[i]))
               break;
         if (i < nmul)
            strcat(out, demul[i]);
         else
            strcat(out, s);
      }
      strcat(out, " ");
   }
}


char      wordbuff[MAXWORD];


char *GetWord(char **p)
{
   while (SpaceChar(**p))
      (*p)++;

   if (**p == 0)
      return NULL;

   if (LetterChar(**p)) {
      char *s = wordbuff;

      while (LetterChar(**p))
         *(s++) = *((*p)++);

      *s = 0;
   } else {
      wordbuff[0] = **p;
      wordbuff[1] = 0;
      (*p)++;
   }

   return wordbuff;
}

BOOL SpaceChar(unsigned char c)
{
	return c==' ';
}


BOOL LetterChar(unsigned char c)
{
	return lextab[c]==ST_CHAR;
}


int enclitic(char *s)
{
   int   i;
   int   k = strlen(s);

   for (i = 0; i < nenclitic; i++) {
      int   kk = strlen(enclit[i]);

      if (!strcmp(s + k - kk, enclit[i]))
         return k - kk;
   }
   return 0;
}
